import shutil

from tornado.httpserver import HTTPServer
from tornado.netutil import (
    bind_sockets,
    bind_unix_socket,
)

from pcs.daemon import log
from pcs.daemon.ssl import PcsdSSL
from pcs.lib.auth.const import ADMIN_GROUP


class HttpsServerManageException(Exception):
    pass


# pylint: disable=too-many-instance-attributes
class HttpsServerManage:
    """
    Instance of HttpsServerManage encapsulates the construction of an HTTPServer
    """

    # Main motivation for this object is to be able to change the ssl
    # certificates from an http request without restarting the whole pcsd
    # daemon.
    #
    # For this purpose an application, which handles http requests, gets
    # a reference to the HttpsServerManage instance. When new certificates
    # arrive via a request the application asks the HttpsServerManage instance
    # for necessary steps (it stops listening of the current HTTPServer and
    # starts a new one with updated certificates).

    def __init__(
        self,
        make_app,
        port,
        bind_addresses,
        ssl: PcsdSSL,
        unix_socket_path,
    ):
        self.__make_app = make_app
        self.__port = port
        self.__bind_addresses = bind_addresses
        self.__tcp_server = None
        self.__ssl = ssl
        self.__unix_socket_path = unix_socket_path
        self.__unix_socket_server = None
        self.__server_is_running = False

    @property
    def server_is_running(self):
        return self.__server_is_running

    def stop(self):
        self.__tcp_server.stop()
        self.__unix_socket_server.stop()
        self.__server_is_running = False

    def start(self):
        self.__ssl.guarantee_valid_certs()

        log.pcsd.info("Starting server...")

        app = self.__make_app(self)
        self.__tcp_server = HTTPServer(
            app,
            ssl_options=self.__ssl.create_context(),
        )
        self.__unix_socket_server = HTTPServer(app)

        # It is necessary to bind sockets for every new HTTPServer since
        # HTTPServer.stop calls sock.close() inside.
        sockets = []
        for address in self.__bind_addresses:
            log.pcsd.info(
                "Binding socket for address '%s' and port '%s'",
                address if address is not None else "*",
                self.__port,
            )
            sockets.extend(bind_sockets(self.__port, address))

        self.__tcp_server.add_sockets(sockets)
        self.__unix_socket_server.add_socket(
            bind_unix_socket(self.__unix_socket_path, mode=0o660)
        )
        shutil.chown(self.__unix_socket_path, 0, ADMIN_GROUP)

        log.pcsd.info("Server is listening")
        self.__server_is_running = True
        return self

    def reload_certs(self):
        if not self.server_is_running:
            raise HttpsServerManageException(
                "Could not reload certificates, server is not running"
            )
        log.pcsd.info("Stopping server to reload ssl certificates...")
        self.stop()
        self.start()



from unittest import TestCase

from lxml import etree

from pcs.common import reports
from pcs.lib.cib.resource import stonith

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
)


class IsStonithEnabled(TestCase):
    def test_not_set(self):
        crm_config = etree.fromstring("<crm_config />")
        self.assertTrue(stonith.is_stonith_enabled(crm_config))

    def test_set_to_enabled(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="abc" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertTrue(stonith.is_stonith_enabled(crm_config))

    def test_set_to_disabled(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="abc" value="true" />
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertFalse(stonith.is_stonith_enabled(crm_config))

    def test_multiple_values(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertFalse(stonith.is_stonith_enabled(crm_config))

    def test_multiple_sections(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                </cluster_property_set>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertFalse(stonith.is_stonith_enabled(crm_config))


class GetMisconfiguredResources(TestCase):
    def test_no_stonith(self):
        resources = etree.fromstring(
            """
            <resources>
                <primitive id="R" class="ocf" provider="pacemaker" type="Dummy">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
            </resources>
        """
        )
        self.assertEqual(
            stonith.get_misconfigured_resources(resources), ([], [], [])
        )

    def test_all_ok(self):
        resources = etree.fromstring(
            """
            <resources>
                <primitive id="S1" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="name" value="value" />
                    </instance_attributes>
                </primitive>
            </resources>
        """
        )
        self.assertEqual(
            stonith.get_misconfigured_resources(resources),
            (
                resources.findall("primitive[@id='S1']"),
                [],
                [],
            ),
        )

    def test_issues(self):
        resources = etree.fromstring(
            """
            <resources>
                <primitive id="S1" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="name" value="value" />
                        <nvpair name="method" value="onoff" />
                    </instance_attributes>
                </primitive>
                <primitive id="S2" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                    </instance_attributes>
                </primitive>
                <primitive id="S3" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
                <primitive id="S4" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
            </resources>
        """
        )
        stonith1 = resources.find("primitive[@id='S1']")
        stonith2 = resources.find("primitive[@id='S2']")
        stonith3 = resources.find("primitive[@id='S3']")
        stonith4 = resources.find("primitive[@id='S4']")
        self.assertEqual(
            stonith.get_misconfigured_resources(resources),
            (
                [stonith1, stonith2, stonith3, stonith4],
                [stonith2, stonith4],
                [stonith3, stonith4],
            ),
        )


class ValidateStonithRestartlessUpdate(TestCase):
    RESOURCES = etree.fromstring(
        """
        <resources>
            <primitive id="supported" class="stonith" type="fence_scsi">
                <instance_attributes>
                    <nvpair name="devices" value="/dev/sda" />
                </instance_attributes>
            </primitive>
            <primitive id="empty" class="stonith" type="fence_scsi">
                <instance_attributes>
                    <nvpair id="empty-instance_attributes-devices"
                        name="devices" value="" />
                </instance_attributes>
            </primitive>
            <primitive id="no-devices" class="stonith" type="fence_scsi"/>
            <primitive id="unsupported_provider"
                class="stonith" provider="provider" type="fence_scsi"/>
            <primitive id="unsupported_type" class="stonith" type="fence_xvm"/>
            <primitive class="ocf" id="cp-01" provider="pacemaker" type="Dummy"/>
        </resources>
        """
    )

    def assert_unsupported_stonith_agent(self, resource_id, resource_type):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, resource_id
        )
        self.assertEqual(
            stonith_el,
            self.RESOURCES.find(f".//primitive[@id='{resource_id}']"),
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNSUPPORTED_AGENT,
                    resource_id=resource_id,
                    resource_type=resource_type,
                    supported_stonith_types=["fence_scsi", "fence_mpath"],
                )
            ],
        )

    def assert_no_devices(self, resource_id):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, resource_id
        )
        self.assertEqual(
            stonith_el,
            self.RESOURCES.find(f".//primitive[@id='{resource_id}']"),
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=(
                        "no devices option configured for stonith device "
                        f"'{resource_id}'"
                    ),
                    reason_type="other",
                )
            ],
        )

    def test_supported(self):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, "supported"
        )
        self.assertEqual(
            stonith_el, self.RESOURCES.find(".//primitive[@id='supported']")
        )
        assert_report_item_list_equal(report_list, [])

    def test_nonexistent_id(self):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, "non-existent"
        )
        self.assertEqual(stonith_el, None)
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="non-existent",
                    expected_types=["primitive"],
                    context_type="resources",
                    context_id="",
                )
            ],
        )

    def test_not_a_resource_id(self):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, "empty-instance_attributes-devices"
        )
        self.assertEqual(stonith_el, None)
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="empty-instance_attributes-devices",
                    expected_types=["primitive"],
                    current_type="nvpair",
                )
            ],
        )

    def test_devices_empty(self):
        self.assert_no_devices("empty")

    def test_missing_devices_attr(self):
        self.assert_no_devices("no-devices")

    def test_unsupported_class(self):
        self.assert_unsupported_stonith_agent("cp-01", "Dummy")

    def test_unsupported_provider(self):
        self.assert_unsupported_stonith_agent(
            "unsupported_provider", "fence_scsi"
        )

    def test_unsupported_type(self):
        self.assert_unsupported_stonith_agent("unsupported_type", "fence_xvm")


def _fixture_stonith_el(host_map_value):
    if host_map_value is None:
        return etree.fromstring(
            '<primitive id="fence-mepath" class="stonith" type="fence_mpath" />'
        )
    return etree.fromstring(
        f"""
        <primitive id="fence-mepath" class="stonith" type="fence_mpath">
            <instance_attributes id="fence-mepath-instance_attributes">
                <nvpair id="fence-mepath-instance_attributes-pcmk_host_map"
                    name="pcmk_host_map" value="{host_map_value}" />
            </instance_attributes>
        </primitive>
        """
    )


class GetNodeKeyMapForMpath(TestCase):
    NODE_LABELS = ["rh9-1", "rh9-2", "rh9-3"]

    def assert_success(self, value, result_dict, node_labels=None):
        if node_labels is None:
            node_labels = self.NODE_LABELS
        self.assertEqual(
            stonith.get_node_key_map_for_mpath(
                _fixture_stonith_el(value), node_labels
            ),
            result_dict,
        )

    def assert_error(self, value, node_labels=None, missing_nodes=None):
        if node_labels is None:
            node_labels = self.NODE_LABELS
        if missing_nodes is None:
            missing_nodes = self.NODE_LABELS
        assert_raise_library_error(
            lambda: stonith.get_node_key_map_for_mpath(
                _fixture_stonith_el(value), node_labels
            ),
            fixture.error(
                reports.codes.STONITH_RESTARTLESS_UPDATE_MISSING_MPATH_KEYS,
                pcmk_host_map_value=value if value else None,
                missing_nodes=sorted(missing_nodes),
            ),
        )

    def test_invalid_part_skipped(self):
        self.assert_success(
            "  ;\trh9-1:1;rh9-2==22;rh9-3 :3;e",
            {"rh9-1": "1"},
            node_labels=self.NODE_LABELS[0:1],
        )

    def test_success_colon_sign_as_assignment_char(self):
        self.assert_success(
            "rh9-1:1;rh9-2:2;rh9-3:3",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3"},
        )

    def test_success_equal_sign_as_assignment_char(self):
        self.assert_success(
            "rh9-1=1;rh9-2=2;rh9-3=3",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3"},
        )

    def test_success_mixed_assignment_char(self):
        self.assert_success(
            "rh9-1:1;rh9-2=2;rh9-3:3",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3"},
        )

    def test_success_semicolon_as_separator(self):
        self.assert_success(
            "rh9-1:1;rh9-2:2;rh9-3:3;",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3"},
        )

    def test_success_space_as_separator(self):
        self.assert_success(
            "rh9-1:1 rh9-2:2 rh9-3:3 ",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3"},
        )

    def test_success_tab_as_separator(self):
        self.assert_success(
            "rh9-1:1\trh9-2:2\trh9-3:3\t",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3"},
        )

    def test_success_mixed_separators(self):
        self.assert_success(
            "rh9-1:1;rh9-2:2 rh9-3:3\trh9-4:4",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3", "rh9-4": "4"},
            node_labels=self.NODE_LABELS + ["rh9-4"],
        )

    def test_success_more_keys_than_current_nodes(self):
        self.assert_success(
            "rh9-1:1;rh9-2:2;rh9-3:3",
            {"rh9-1": "1", "rh9-2": "2", "rh9-3": "3"},
            node_labels=self.NODE_LABELS[:-1],
        )

    def test_empty_nodes(self):
        self.assert_success(
            "rh9-1:1;rh9-2:2", {"rh9-1": "1", "rh9-2": "2"}, node_labels=[]
        )

    def test_missing_value(self):
        self.assert_error(None)

    def test_empty_value(self):
        self.assert_error("")

    def test_not_enough_keys_for_nodes_misconfiguration(self):
        self.assert_error("rh9-1:1;rh9-2;rh9-3:3", missing_nodes=["rh9-2"])

    def test_not_enough_keys_for_nodes_missing(self):
        self.assert_error("rh9-1:1", missing_nodes=["rh9-3", "rh9-2"])



import os
import os.path
import re
import shutil
from typing import (
    Callable,
    Optional,
    Sequence,
)

from pcs import settings
from pcs.common import reports
from pcs.common.node_communicator import (
    NodeCommunicatorFactory,
    RequestTarget,
)
from pcs.common.str_tools import join_multilines
from pcs.common.types import StringSequence
from pcs.lib.communication import qdevice_net as qdevice_net_com
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.tools import get_tmp_file

SERVICE_NAME = "corosync-qnetd"

__model = "net"
__qnetd_certutil = os.path.join(
    settings.corosync_qnet_binaries, "corosync-qnetd-certutil"
)
__qnetd_tool = os.path.join(
    settings.corosync_qnet_binaries, "corosync-qnetd-tool"
)
__qdevice_certutil = os.path.join(
    settings.corosync_qdevice_binaries, "corosync-qdevice-net-certutil"
)
__nss_certificate_db_files = (
    # new NSS DB format
    "cert9.db",
    "key4.db",
    "pkcs11.txt",
    # old NSS DB format
    "cert8.db",
    "key3.db",
    "secmod.db",
)


class QnetdNotRunningException(Exception):
    pass


def set_up_client_certificates(
    runner: CommandRunner,
    reporter: reports.ReportProcessor,
    communicator_factory: NodeCommunicatorFactory,
    qnetd_target: RequestTarget,
    cluster_name: str,
    cluster_nodes_target_list: Sequence[RequestTarget],
    skip_offline_nodes: bool,
    allow_skip_offline: bool = True,
) -> None:
    """
    setup cluster nodes for using qdevice model net

    runner -- command runner instance
    reporter -- report processor instance
    communicator_factory -- communicator facto. instance
    qnetd_target -- qdevice provider (qnetd host)
    cluster_name -- name of the cluster to which qdevice is being added
    cluster_nodes_target_list -- list of cluster nodes targets
    skip_offline_nodes -- continue even if not all nodes are accessible
    allow_skip_offline -- enables forcing errors by skip_offline_nodes
    """
    # pylint: disable=too-many-locals
    reporter.report(
        reports.ReportItem.info(
            reports.messages.QdeviceCertificateDistributionStarted()
        )
    )
    # get qnetd CA certificate
    com_cmd_1 = qdevice_net_com.GetCaCert(reporter)
    com_cmd_1.set_targets([qnetd_target])
    qnetd_ca_cert = run_and_raise(
        communicator_factory.get_communicator(), com_cmd_1
    )[0][1]
    # init certificate storage on all nodes
    com_cmd_2 = qdevice_net_com.ClientSetup(
        reporter, qnetd_ca_cert, skip_offline_nodes, allow_skip_offline
    )
    com_cmd_2.set_targets(cluster_nodes_target_list)
    run_and_raise(communicator_factory.get_communicator(), com_cmd_2)
    # create client certificate request
    cert_request = client_generate_certificate_request(runner, cluster_name)
    # sign the request on qnetd host
    com_cmd_3 = qdevice_net_com.SignCertificate(reporter)
    com_cmd_3.add_request(qnetd_target, cert_request, cluster_name)
    signed_certificate = run_and_raise(
        communicator_factory.get_communicator(), com_cmd_3
    )[0][1]
    # transform the signed certificate to pk12 format which can sent to nodes
    pk12 = client_cert_request_to_pk12(runner, signed_certificate)
    # distribute final certificate to nodes
    com_cmd_4 = qdevice_net_com.ClientImportCertificateAndKey(
        reporter, pk12, skip_offline_nodes, allow_skip_offline
    )
    com_cmd_4.set_targets(cluster_nodes_target_list)
    run_and_raise(communicator_factory.get_communicator(), com_cmd_4)


def qdevice_setup(runner: CommandRunner) -> None:
    """
    initialize qdevice on local host

    runner -- command runner instance
    """
    if qdevice_initialized():
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceAlreadyInitialized(__model)
            )
        )

    stdout, stderr, retval = runner.run([__qnetd_certutil, "-i"])
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )


def qdevice_initialized() -> bool:
    """
    check if qdevice server certificate database has been initialized
    """
    return _nss_certificate_db_initialized(
        settings.corosync_qdevice_net_server_certs_dir
    )


def qdevice_destroy() -> None:
    """
    delete qdevice configuration on local host
    """
    try:
        if qdevice_initialized():
            shutil.rmtree(settings.corosync_qdevice_net_server_certs_dir)
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceDestroyError(__model, e.strerror)
            )
        ) from e


def qdevice_status_generic_text(
    runner: CommandRunner, verbose: bool = False
) -> str:
    """
    get qdevice runtime status in plain text

    verbose -- get more detailed output
    """
    args = ["-s"]
    if verbose:
        args.append("-v")
    stdout, stderr, retval = _qdevice_run_tool(runner, args)
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceGetStatusError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout


def qdevice_status_cluster_text(
    runner: CommandRunner, cluster: Optional[str] = None, verbose: bool = False
) -> str:
    """
    get qdevice runtime status in plain text

    cluster -- show information only about specified cluster
    verbose -- get more detailed output
    """
    args = ["-l"]
    if verbose:
        args.append("-v")
    if cluster:
        args.extend(["-c", cluster])
    stdout, stderr, retval = _qdevice_run_tool(runner, args)
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceGetStatusError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return stdout


def qdevice_connected_clusters(status_cluster_text: str) -> list[str]:
    """
    parse qnetd cluster status listing and return connected clusters' names

    status_cluster_text -- output of corosync-qnetd-tool -l
    """
    connected_clusters = []
    regexp = re.compile(r'^Cluster "(?P<cluster>[^"]+)":$')
    for line in status_cluster_text.splitlines():
        match = regexp.search(line)
        if match:
            connected_clusters.append(match.group("cluster"))
    return connected_clusters


def _qdevice_run_tool(
    runner: CommandRunner, args: StringSequence
) -> tuple[str, str, int]:
    """
    run corosync-qnetd-tool, raise QnetdNotRunningException if qnetd not running

    iterable args -- corosync-qnetd-tool arguments
    """
    stdout, stderr, retval = runner.run([__qnetd_tool] + list(args))
    if retval == 3 and "is qnetd running?" in stderr.lower():
        raise QnetdNotRunningException()
    return stdout, stderr, retval


def qdevice_sign_certificate_request(
    runner: CommandRunner, cert_request: bytes, cluster_name: str
) -> bytes:
    """
    sign client certificate request

    cert_request -- certificate request data
    cluster_name -- name of the cluster to which qdevice is being added
    """
    if not qdevice_initialized():
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceNotInitialized(__model)
            )
        )
    # save the certificate request, corosync tool only works with files
    try:
        with get_tmp_file(cert_request, binary=True) as tmpfile:
            # sign the request
            stdout, stderr, retval = runner.run(
                [__qnetd_certutil, "-s", "-c", tmpfile.name, "-n", cluster_name]
            )
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateSignError(e.strerror)
            )
        ) from e
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateSignError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    # get signed certificate, corosync tool only works with files
    return _get_output_certificate(
        stdout,
        reports.messages.QdeviceCertificateSignError,
    )


def client_setup(runner: CommandRunner, ca_certificate: bytes) -> None:
    """
    initialize qdevice client on local host

    ca_certificate -- qnetd CA certificate
    """
    client_destroy()
    # save CA certificate, corosync tool only works with files
    try:
        if not os.path.exists(settings.corosync_qdevice_net_client_certs_dir):
            os.makedirs(
                settings.corosync_qdevice_net_client_certs_dir, mode=0o700
            )
        ca_file_path = os.path.join(
            settings.corosync_qdevice_net_client_certs_dir,
            settings.corosync_qdevice_net_client_ca_file_name,
        )
        with open(ca_file_path, "wb") as ca_file:
            ca_file.write(ca_certificate)
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    e.strerror,
                )
            )
        ) from e
    # initialize client's certificate storage
    stdout, stderr, retval = runner.run(
        [__qdevice_certutil, "-i", "-c", ca_file_path]
    )
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )


def client_initialized() -> bool:
    """
    check if qdevice net client certificate database has been initialized
    """
    return _nss_certificate_db_initialized(
        settings.corosync_qdevice_net_client_certs_dir
    )


def client_destroy() -> None:
    """
    delete qdevice client config files on local host
    """
    try:
        if client_initialized():
            shutil.rmtree(settings.corosync_qdevice_net_client_certs_dir)
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceDestroyError(__model, e.strerror)
            )
        ) from e


def client_generate_certificate_request(
    runner: CommandRunner, cluster_name: str
) -> bytes:
    """
    create a certificate request which can be signed by qnetd server

    cluster_name -- name of the cluster to which qdevice is being added
    """
    if not client_initialized():
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceNotInitialized(__model)
            )
        )
    stdout, stderr, retval = runner.run(
        [__qdevice_certutil, "-r", "-n", cluster_name]
    )
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceInitializationError(
                    __model,
                    join_multilines([stderr, stdout]),
                )
            )
        )
    return _get_output_certificate(
        stdout,
        lambda reason: reports.messages.QdeviceInitializationError(
            __model,
            reason,
        ),
    )


def client_cert_request_to_pk12(
    runner: CommandRunner, cert_request: bytes
) -> bytes:
    """
    transform signed certificate request to pk12 certificate which can be
    imported to nodes

    cert_request -- signed certificate request
    """
    if not client_initialized():
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceNotInitialized(__model)
            )
        )
    # save the signed certificate request, corosync tool only works with files
    try:
        with get_tmp_file(cert_request, binary=True) as tmpfile:
            # transform it
            stdout, stderr, retval = runner.run(
                [__qdevice_certutil, "-M", "-c", tmpfile.name]
            )
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateImportError(e.strerror)
            )
        ) from e
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateImportError(
                    join_multilines([stderr, stdout]),
                )
            )
        )
    # get resulting pk12, corosync tool only works with files
    return _get_output_certificate(
        stdout, reports.messages.QdeviceCertificateImportError
    )


def client_import_certificate_and_key(
    runner: CommandRunner, pk12_certificate: bytes
) -> None:
    """
    import qdevice client certificate to the local node certificate storage
    """
    if not client_initialized():
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceNotInitialized(__model)
            )
        )
    # save the certificate, corosync tool only works with files
    try:
        with get_tmp_file(pk12_certificate, binary=True) as tmpfile:
            stdout, stderr, retval = runner.run(
                [__qdevice_certutil, "-m", "-c", tmpfile.name]
            )
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateImportError(e.strerror)
            )
        ) from e
    if retval != 0:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.QdeviceCertificateImportError(
                    join_multilines([stderr, stdout]),
                )
            )
        )


def _nss_certificate_db_initialized(cert_db_path: str) -> bool:
    for filename in __nss_certificate_db_files:
        if os.path.exists(os.path.join(cert_db_path, filename)):
            return True
    return False


def _get_output_certificate(
    cert_tool_output: str,
    report_message_func: Callable[[str], reports.ReportItemMessage],
) -> bytes:
    regexp = re.compile(r"^Certificate( request)? stored in (?P<path>.+)$")
    filename = None
    for line in cert_tool_output.splitlines():
        match = regexp.search(line)
        if match:
            filename = match.group("path")
    if not filename:
        raise LibraryError(
            reports.ReportItem.error(report_message_func(cert_tool_output))
        )
    try:
        with open(filename, "rb") as cert_file:
            return cert_file.read()
    except EnvironmentError as e:
        raise LibraryError(
            reports.ReportItem.error(
                report_message_func(
                    "{path}: {error}".format(path=filename, error=e.strerror)
                )
            )
        ) from e



from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    InputModifiers,
    ensure_unique_args,
    prepare_options,
)
from pcs.common import reports
from pcs.common.types import StringSequence


def set_property(
    lib: Any, argv: StringSequence, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow unknown options
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    if not argv:
        raise CmdLineInputError()
    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)
    cluster_options = prepare_options(argv)
    lib.cluster_property.set_properties(cluster_options, force_flags)


def unset_property(
    lib: Any, argv: StringSequence, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - no error when removing not existing properties
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    if not argv:
        raise CmdLineInputError()
    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)
    else:
        ensure_unique_args(argv)

    lib.cluster_property.set_properties(
        {name: "" for name in argv}, force_flags
    )



from collections.abc import Set
from enum import (
    Enum,
    auto,
)
from typing import (
    Generator,
    MutableSequence,
    Union,
)

StringSequence = Union[MutableSequence[str], tuple[str, ...]]
StringCollection = Union[StringSequence, Set[str]]
StringIterable = Union[StringCollection, Generator[str, None, None]]


class AutoNameEnum(str, Enum):
    def _generate_next_value_(name, start, count, last_values):  # type: ignore
        # pylint: disable=no-self-argument
        del start, count, last_values
        return name


class CibRuleExpressionType(AutoNameEnum):
    RULE = auto()
    EXPRESSION = auto()  # node attribute expression, named 'expression' in CIB
    DATE_EXPRESSION = auto()
    OP_EXPRESSION = auto()
    RSC_EXPRESSION = auto()


class CibRuleInEffectStatus(AutoNameEnum):
    NOT_YET_IN_EFFECT = auto()
    IN_EFFECT = auto()
    EXPIRED = auto()
    UNKNOWN = auto()


class ResourceRelationType(AutoNameEnum):
    ORDER = auto()
    ORDER_SET = auto()
    INNER_RESOURCES = auto()
    OUTER_RESOURCE = auto()
    RSC_PRIMITIVE = auto()
    RSC_CLONE = auto()
    RSC_GROUP = auto()
    RSC_BUNDLE = auto()
    RSC_UNKNOWN = auto()


class DrRole(AutoNameEnum):
    PRIMARY = auto()
    RECOVERY = auto()


class UnknownCorosyncTransportTypeException(Exception):
    def __init__(self, transport: str) -> None:
        super().__init__()
        self.transport = transport


class CorosyncTransportType(AutoNameEnum):
    UDP = auto()
    UDPU = auto()
    KNET = auto()

    @classmethod
    def from_str(cls, transport: str) -> "CorosyncTransportType":
        try:
            return cls(transport.upper())
        except ValueError:
            raise UnknownCorosyncTransportTypeException(transport) from None


class CorosyncNodeAddressType(Enum):
    IPV4 = "IPv4"
    IPV6 = "IPv6"
    FQDN = "FQDN"
    UNRESOLVABLE = "unresolvable"



from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs import settings
from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import resource

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc


def _node_fixture(name, node_id):
    return f'<node id="{node_id}" uname="{name}"/>'


def _node_list_fixture(nodes):
    return "\n".join(
        _node_fixture(node_name, node_id)
        for node_id, node_name in enumerate(nodes)
    )


def _nodes_section_fixture(content):
    return f"""
    <nodes>
    {content}
    </nodes>
    """


nodes_section = _nodes_section_fixture(
    _node_list_fixture(["node", "node1", "node2"])
)
resources_primitive = """
    <resources>
        <primitive id="A" />
    </resources>
"""
resources_promotable = """
    <resources>
        <clone id="A-clone">
            <primitive id="A" />
            <meta_attributes>
                <nvpair name="promotable" value="true" />
            </meta_attributes>
        </clone>
    </resources>
"""


def _action_kwargs(kwargs):
    defaulted = dict(
        resource="A",
        stdout="pcmk std out",
        stderr="pcmk std err",
        returncode=0,
    )
    defaulted.update(kwargs)
    return defaulted


class MoveBanClearBaseMixin:
    # pylint does not and can not know setUp is a method of TestCase
    # pylint: disable=invalid-name
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_resource_not_found(self):
        self.config.runner.cib.load(resources=resources_primitive)
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(self.env_assist.get_env(), "B")
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found("B", context_type="resources"),
            ]
        )

    def test_master_of_nonpromotable_resource(self):
        # This is a basic test which checks validation is being done. It
        # utilizes the fact all three commands (move, ban, unmove_unban) share
        # a common situation to check. Detailed validation tests are in
        # pcs_test.tier0.lib.cib.test_resource_common
        self.config.runner.cib.load(resources=resources_primitive)
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(self.env_assist.get_env(), "A", master=True)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="A",
                    promotable_id=None,
                ),
            ]
        )

    def test_pcmk_error(self):
        self.config.runner.cib.load(resources=resources_primitive)
        self.config_pcmk_action(returncode=1)
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(self.env_assist.get_env(), "A"),
            [
                fixture.error(
                    self.report_code_pcmk_error,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
            ],
            expected_in_processor=False,
        )

    def test_success(self):
        self.config.runner.cib.load(resources=resources_primitive)
        self.config_pcmk_action()
        self.lib_action(self.env_assist.get_env(), "A")
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
            ]
        )


class MoveBanBaseMixin(MoveBanClearBaseMixin):
    def test_resource_stopped(self):
        stderr = dedent(
            """\
            Resource 'A' not moved: active in 0 locations.
            To prevent 'A' from running on a specific location, specify a node.
            Error performing operation: Invalid argument
            """
        )
        self.config.runner.cib.load(resources=resources_primitive)
        self.config_pcmk_action(stderr=stderr, returncode=1)
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(self.env_assist.get_env(), "A"),
            [
                fixture.error(
                    self.report_code_resource_stopped,
                    resource_id="A",
                ),
            ],
            expected_in_processor=False,
        )

    def test_node_not_found(self):
        self.config.runner.cib.load(resources=resources_primitive)
        node = "node"
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(self.env_assist.get_env(), "A", node)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND, node=node, searched_types=[]
                )
            ]
        )

    def test_all_options(self):
        self.config.runner.cib.load(
            resources=resources_promotable, nodes=nodes_section
        )
        self.config_pcmk_action(
            resource="A-clone",
            master=True,
            node="node",
            lifetime="1h",
        )
        self.lib_action(
            self.env_assist.get_env(),
            "A-clone",
            master=True,
            node="node",
            lifetime="1h",
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A-clone",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
            ]
        )


class MoveMixin:
    lib_action = staticmethod(resource.move)
    report_code_bad_master = (
        report_codes.CANNOT_MOVE_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )
    report_code_pcmk_error = report_codes.RESOURCE_MOVE_PCMK_ERROR
    report_code_pcmk_success = report_codes.RESOURCE_MOVE_PCMK_SUCCESS
    report_code_resource_stopped = (
        report_codes.CANNOT_MOVE_RESOURCE_STOPPED_NO_NODE_SPECIFIED
    )

    def config_pcmk_action(self, *args, **kwargs):
        self.config.runner.pcmk.resource_move(*args, **_action_kwargs(kwargs))


class Move(MoveMixin, MoveBanBaseMixin, TestCase):
    def test_resource_may_or_may_not_move(self):
        self.config.runner.cib.load(
            resources=resources_primitive, nodes=nodes_section
        )
        self.config_pcmk_action(
            resource="A", node="node1", stdout="", stderr=""
        )
        self.lib_action(self.env_assist.get_env(), "A", node="node1")
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.RESOURCE_MAY_OR_MAY_NOT_MOVE,
                    resource_id="A",
                ),
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="",
                    stderr="",
                ),
            ]
        )


class BanMixin:
    lib_action = staticmethod(resource.ban)
    report_code_bad_master = (
        report_codes.CANNOT_BAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )
    report_code_pcmk_error = report_codes.RESOURCE_BAN_PCMK_ERROR
    report_code_pcmk_success = report_codes.RESOURCE_BAN_PCMK_SUCCESS
    report_code_resource_stopped = (
        report_codes.CANNOT_BAN_RESOURCE_STOPPED_NO_NODE_SPECIFIED
    )

    def config_pcmk_action(self, *args, **kwargs):
        self.config.runner.pcmk.resource_ban(*args, **_action_kwargs(kwargs))


class Ban(BanMixin, MoveBanBaseMixin, TestCase):
    pass


class UnmoveUnbanMixin:
    lib_action = staticmethod(resource.unmove_unban)
    report_code_bad_master = (
        report_codes.CANNOT_UNMOVE_UNBAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )
    report_code_pcmk_error = report_codes.RESOURCE_UNMOVE_UNBAN_PCMK_ERROR
    report_code_pcmk_success = report_codes.RESOURCE_UNMOVE_UNBAN_PCMK_SUCCESS

    def config_pcmk_action(self, *args, **kwargs):
        self.config.runner.pcmk.resource_clear(*args, **_action_kwargs(kwargs))


class UnmoveUnban(UnmoveUnbanMixin, MoveBanClearBaseMixin, TestCase):
    def test_all_options(self):
        self.config.runner.cib.load(resources=resources_promotable)
        self.config.runner.pcmk.can_clear_expired()
        self.config_pcmk_action(
            resource="A-clone", master=True, node="node", expired=True
        )
        self.lib_action(
            self.env_assist.get_env(),
            "A-clone",
            master=True,
            node="node",
            expired=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A-clone",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
            ]
        )

    def test_expired_not_supported(self):
        self.config.runner.cib.load(resources=resources_promotable)
        self.config.runner.pcmk.can_clear_expired(stderr="not supported")
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", expired=True
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    report_codes.RESOURCE_UNMOVE_UNBAN_PCMK_EXPIRED_NOT_SUPPORTED
                ),
            ]
        )


class MoveBanWaitMixin:
    state_running_node1 = """
        <resources>
            <resource id="A" role="Started" nodes_running_on="1">
                 <node name="node1" id="1" cached="false" />
             </resource>
        </resources>
    """
    state_running_node2 = """
        <resources>
            <resource id="A" role="Started" nodes_running_on="1">
                 <node name="node2" id="2" cached="false" />
             </resource>
        </resources>
    """
    state_not_running = """
        <resources>
            <resource id="A" role="Stopped" nodes_running_on="0">
             </resource>
        </resources>
    """
    # pylint does not and can not know setUp is a method of TestCase
    # pylint: disable=invalid-name
    def setUp(self):
        self.timeout = 10
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(
            resources=resources_primitive, nodes=nodes_section
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_state_before_action_fail(self):
        self.config.runner.pcmk.load_state(
            stdout="state stdout", stderr="state stderr", returncode=1
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(self.env_assist.get_env(), "A", wait="10"),
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason="state stderr\nstate stdout",
                ),
            ],
            expected_in_processor=False,
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_state_before_action_not_xml(self):
        self.config.runner.pcmk.load_state(stdout="state stdout")
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=str(self.timeout)
            ),
            [
                fixture.error(
                    report_codes.BAD_CLUSTER_STATE_FORMAT,
                ),
            ],
            expected_in_processor=False,
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_pcmk_fail(self):
        self.config.runner.pcmk.load_state()
        self.config_pcmk_action(returncode=1)
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=str(self.timeout)
            ),
            [
                fixture.error(
                    self.report_code_pcmk_error,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
            ],
            expected_in_processor=False,
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_fail(self):
        self.config.runner.pcmk.load_state()
        self.config_pcmk_action()
        self.config.runner.pcmk.wait(
            timeout=self.timeout, stderr="wait error", returncode=1
        )

        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=str(self.timeout)
            ),
            [
                fixture.error(
                    report_codes.WAIT_FOR_IDLE_ERROR,
                    reason="wait error",
                ),
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_state_after_action_fail(self):
        self.config.runner.pcmk.load_state()
        self.config_pcmk_action()
        self.config.runner.pcmk.wait(timeout=self.timeout)
        self.config.runner.pcmk.load_state(
            name="runner.pcmk.load_state.after",
            stdout="state stdout",
            stderr="state stderr",
            returncode=1,
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=self.timeout
            ),
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason="state stderr\nstate stdout",
                ),
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_state_after_action_not_xml(self):
        self.config.runner.pcmk.load_state()
        self.config_pcmk_action()
        self.config.runner.pcmk.wait(timeout=self.timeout)
        self.config.runner.pcmk.load_state(
            name="runner.pcmk.load_state.after", stdout="state stdout"
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=self.timeout
            ),
            [
                fixture.error(
                    report_codes.BAD_CLUSTER_STATE_FORMAT,
                ),
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
            ]
        )

    def test_was_running_now_stopped(self):
        self.success_config(self.state_running_node1, self.state_not_running)
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=self.timeout
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_not_running(
                    "A", severity=severities.ERROR
                ),
            ]
        )

    def test_was_stopped_now_stopped(self):
        self.success_config(self.state_not_running, self.state_not_running)
        self.lib_action(self.env_assist.get_env(), "A", wait=self.timeout)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_not_running(
                    "A",
                ),
            ]
        )

    def test_running_on_same_node_no_node_specified(self):
        self.success_config(self.state_running_node1, self.state_running_node1)
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=self.timeout
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_running(
                    "A", {"Started": ["node1"]}, severity=severities.ERROR
                ),
            ]
        )

    def test_running_on_onther_node_no_node_specified(self):
        self.success_config(self.state_running_node1, self.state_running_node2)
        self.lib_action(self.env_assist.get_env(), "A", wait=self.timeout)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_running(
                    "A",
                    {"Started": ["node2"]},
                ),
            ]
        )

    def success_config(self, state_before, state_after, action_node=None):
        self.config.runner.pcmk.load_state(resources=state_before)
        self.config_pcmk_action(node=action_node)
        self.config.runner.pcmk.wait(timeout=self.timeout)
        self.config.runner.pcmk.load_state(
            name="runner.pcmk.load_state.after",
            resources=state_after,
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class MoveWait(MoveMixin, MoveBanWaitMixin, TestCase):
    def test_running_on_specified_node(self):
        self.success_config(
            self.state_running_node1,
            self.state_running_node2,
            action_node="node2",
        )
        self.lib_action(
            self.env_assist.get_env(), "A", node="node2", wait=self.timeout
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_running(
                    "A",
                    {"Started": ["node2"]},
                ),
            ]
        )

    def test_running_on_not_specified_node(self):
        self.success_config(
            self.state_running_node1,
            self.state_running_node1,
            action_node="node2",
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", node="node2", wait=self.timeout
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_running(
                    "A", {"Started": ["node1"]}, severity=severities.ERROR
                ),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class BanWait(BanMixin, MoveBanWaitMixin, TestCase):
    def test_running_on_specified_node(self):
        self.success_config(
            self.state_running_node1,
            self.state_running_node1,
            action_node="node1",
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", node="node1", wait=self.timeout
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_running(
                    "A", {"Started": ["node1"]}, severity=severities.ERROR
                ),
            ]
        )

    def test_running_on_not_specified_node(self):
        self.success_config(
            self.state_running_node1,
            self.state_running_node2,
            action_node="node1",
        )
        self.lib_action(
            self.env_assist.get_env(), "A", node="node1", wait=self.timeout
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_running(
                    "A",
                    {"Started": ["node2"]},
                ),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class UnmoveUnbanWait(UnmoveUnbanMixin, TestCase):
    def setUp(self):
        self.timeout = 10
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(resources=resources_primitive)
        self.config_pcmk_action()

    def test_success(self):
        resources_state = """
            <resources>
                <resource id="A" role="Started" nodes_running_on="1">
                     <node name="node1" id="1" cached="false" />
                 </resource>
            </resources>
        """
        self.config.runner.pcmk.wait(timeout=self.timeout)
        self.config.runner.pcmk.load_state(resources=resources_state)

        self.lib_action(self.env_assist.get_env(), "A", wait=self.timeout)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
                fixture.report_resource_running("A", {"Started": ["node1"]}),
            ]
        )

    def test_get_state_fail(self):
        self.config.runner.pcmk.wait(timeout=self.timeout)
        self.config.runner.pcmk.load_state(
            stdout="state stdout", stderr="state stderr", returncode=1
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=self.timeout
            ),
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason="state stderr\nstate stdout",
                ),
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
            ]
        )

    def test_get_state_not_xml(self):
        self.config.runner.pcmk.wait(timeout=self.timeout)
        self.config.runner.pcmk.load_state(stdout="state stdout")
        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=self.timeout
            ),
            [
                fixture.error(
                    report_codes.BAD_CLUSTER_STATE_FORMAT,
                ),
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
            ]
        )

    def test_wait_fail(self):
        self.config.runner.pcmk.wait(
            timeout=self.timeout, stderr="wait error", returncode=1
        )

        self.env_assist.assert_raise_library_error(
            lambda: self.lib_action(
                self.env_assist.get_env(), "A", wait=self.timeout
            ),
            [
                fixture.error(
                    report_codes.WAIT_FOR_IDLE_ERROR,
                    reason="wait error",
                ),
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    self.report_code_pcmk_success,
                    resource_id="A",
                    stdout="pcmk std out",
                    stderr="pcmk std err",
                ),
                fixture.info(
                    report_codes.WAIT_FOR_IDLE_STARTED,
                    timeout=self.timeout,
                ),
            ]
        )



# pylint: disable=protected-access
from datetime import timedelta
from unittest import (
    TestCase,
    mock,
)

import pcs.daemon.async_tasks.task as tasks
from pcs.common.async_tasks import types
from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
)
from pcs.common.reports import ReportItemDto
from pcs.daemon.async_tasks.types import Command
from pcs.daemon.async_tasks.worker.types import (
    Message,
    TaskExecuted,
    TaskFinished,
)
from pcs.settings import (
    task_abandoned_timeout_seconds,
    task_unresponsive_timeout_seconds,
)

from .helpers import (
    AUTH_USER,
    DATETIME_NOW,
    MockDateTimeNowMixin,
    MockOsKillMixin,
)

TASK_IDENT = "id0"
TEST_TIMEOUT_S = 10
DATETIME_BEFORE_TIMEOUT = DATETIME_NOW - timedelta(seconds=TEST_TIMEOUT_S / 2)
DATETIME_AFTER_TIMEOUT = DATETIME_NOW - timedelta(seconds=TEST_TIMEOUT_S + 1)
WORKER_PID = 2222


class TaskBaseTestCase(TestCase):
    def setUp(self):
        self.task = tasks.Task(
            TASK_IDENT,
            Command(
                CommandDto(
                    "command", {}, CommandOptionsDto(request_timeout=None)
                )
            ),
            AUTH_USER,
            tasks.TaskConfig(),
        )


class TestReceiveMessage(MockDateTimeNowMixin, TaskBaseTestCase):
    def setUp(self):
        super().setUp()
        self.mock_datetime_now = self._init_mock_datetime_now()

    def test_report(self):
        payload = mock.MagicMock(ReportItemDto)
        message = Message(TASK_IDENT, payload)
        self.task.receive_message(message)
        self.assertEqual([payload], self.task.to_dto().reports)
        self.assertEqual(DATETIME_NOW, self.task._last_message_at)
        self.mock_datetime_now.assert_called_once()

    def test_task_executed(self):
        message = Message(TASK_IDENT, TaskExecuted(WORKER_PID))
        self.task.receive_message(message)
        self.assertEqual(types.TaskState.EXECUTED, self.task.state)
        self.assertEqual(WORKER_PID, self.task._worker_pid)
        self.assertEqual(DATETIME_NOW, self.task._last_message_at)
        self.mock_datetime_now.assert_called()

    def test_task_finished(self):
        message = Message(
            TASK_IDENT,
            TaskFinished(types.TaskFinishType.SUCCESS, "result"),
        )
        self.task.receive_message(message)
        task_dto = self.task.to_dto()
        self.assertEqual(types.TaskState.FINISHED, task_dto.state)
        self.assertEqual(
            types.TaskFinishType.SUCCESS, task_dto.task_finish_type
        )
        self.assertEqual("result", task_dto.result)
        self.assertEqual(DATETIME_NOW, self.task._last_message_at)
        self.mock_datetime_now.assert_called_once()

    def test_unsupported_message_type(self):
        message = Message(TASK_IDENT, 3)
        with self.assertRaises(tasks.UnknownMessageError) as thrown_exc:
            self.task.receive_message(message)
        self.assertEqual(type(3).__name__, thrown_exc.exception.payload_type)
        self.mock_datetime_now.assert_not_called()


class TestRequestKill(TaskBaseTestCase):
    def test_kill_requested(self):
        self.task.request_kill(types.TaskKillReason.USER)
        self.assertEqual(
            types.TaskKillReason.USER, self.task.to_dto().kill_reason
        )
        self.assertTrue(self.task.is_kill_requested())

    def test_kill_not_requested(self):
        self.assertFalse(self.task.is_kill_requested())


class TestKill(MockOsKillMixin, TaskBaseTestCase):
    def setUp(self):
        super().setUp()
        self.mock_os_kill = self._init_mock_os_kill()

    def _assert_killed(self, start_state):
        if self.task.state != start_state:
            self.task.state = start_state
        self.task.kill()
        task_dto = self.task.to_dto()
        self.mock_os_kill.assert_not_called()
        self.assertEqual(types.TaskState.FINISHED, task_dto.state)
        self.assertEqual(types.TaskFinishType.KILL, task_dto.task_finish_type)

    def _assert_not_killed(self, start_state):
        finish_type = self.task.to_dto().task_finish_type
        self.task.state = start_state
        self.task.kill()
        task_dto = self.task.to_dto()
        self.mock_os_kill.assert_not_called()
        self.assertEqual(start_state, task_dto.state)
        self.assertEqual(finish_type, task_dto.task_finish_type)

    def test_kill_created(self):
        self._assert_killed(types.TaskState.CREATED)

    def test_kill_queued(self):
        self._assert_not_killed(types.TaskState.QUEUED)

    def test_kill_executed_worker_alive(self):
        message = Message(TASK_IDENT, TaskExecuted(WORKER_PID))
        self.task.receive_message(message)
        self.task.kill()
        task_dto = self.task.to_dto()
        self.mock_os_kill.assert_called_once_with(WORKER_PID, 15)
        self.assertEqual(types.TaskState.FINISHED, task_dto.state)
        self.assertEqual(types.TaskFinishType.KILL, task_dto.task_finish_type)

    def test_kill_executed_worker_dead(self):
        message = Message(TASK_IDENT, TaskExecuted(WORKER_PID))
        self.task.receive_message(message)
        self.mock_os_kill.raiseError.side_effect = ProcessLookupError()
        self.task.kill()
        task_dto = self.task.to_dto()
        self.mock_os_kill.assert_called_once_with(WORKER_PID, 15)
        self.assertEqual(types.TaskState.FINISHED, task_dto.state)
        self.assertEqual(types.TaskFinishType.KILL, task_dto.task_finish_type)

    def test_kill_finished(self):
        self._assert_not_killed(types.TaskState.FINISHED)


class TestGetLastTimestamp(MockDateTimeNowMixin, TaskBaseTestCase):
    def test_no_messages_created(self):
        self.assertIsNone(self.task._get_last_updated_timestamp())

    def test_no_messages_queued(self):
        self.task.state = types.TaskState.QUEUED
        self.assertIsNone(self.task._get_last_updated_timestamp())

    def test_no_messages_executed(self):
        # This can't happen - task is switched to executed by receiving
        # a TaskExecuted message
        pass

    def test_no_messages_finished(self):
        # This can happen when Task is killed in created state
        self.task.state = types.TaskState.FINISHED
        self._init_mock_datetime_now()
        self.assertEqual(DATETIME_NOW, self.task._get_last_updated_timestamp())
        self.assertEqual(DATETIME_NOW, self.task._last_message_at)

    # There can't be created or queued tasks that have messages, task is
    # assigned to a worker (which sends messages) in executed state
    def test_created(self):
        pass

    def test_queued(self):
        pass

    def test_executed(self):
        self._init_mock_datetime_now()
        message = Message(TASK_IDENT, TaskExecuted(WORKER_PID))
        self.task.receive_message(message)
        self.assertEqual(DATETIME_NOW, self.task._get_last_updated_timestamp())

    def test_finished(self):
        self._init_mock_datetime_now()
        message = Message(
            TASK_IDENT, TaskFinished(types.TaskFinishType.FAIL, None)
        )
        self.task.receive_message(message)
        self.assertEqual(DATETIME_NOW, self.task._get_last_updated_timestamp())


@mock.patch.object(tasks.Task, "_get_last_updated_timestamp")
class TestIsTimedOut(MockDateTimeNowMixin, TaskBaseTestCase):
    def setUp(self):
        super().setUp()
        self._init_mock_datetime_now()  # assertions not needed

    def test_not_timed_out(self, mock_last_timestamp):
        mock_last_timestamp.return_value = DATETIME_BEFORE_TIMEOUT
        self.assertFalse(self.task._is_timed_out(TEST_TIMEOUT_S))
        mock_last_timestamp.assert_called_once()

    def test_timed_out(self, mock_last_timestamp):
        mock_last_timestamp.return_value = DATETIME_AFTER_TIMEOUT
        self.assertTrue(self.task._is_timed_out(TEST_TIMEOUT_S))
        mock_last_timestamp.assert_called_once()


@mock.patch.object(tasks.Task, "_is_timed_out")
class TestDefunct(TaskBaseTestCase):
    def _assert_not_defunct(self, mock_is_timed_out, state):
        if self.task.state != state:
            self.task.state = state
        self.assertFalse(self.task.is_defunct())
        mock_is_timed_out.assert_not_called()

    def test_not_defunct_created(self, mock_is_timed_out):
        self._assert_not_defunct(mock_is_timed_out, types.TaskState.CREATED)

    def test_not_defunct_queued(self, mock_is_timed_out):
        self._assert_not_defunct(mock_is_timed_out, types.TaskState.QUEUED)

    def test_not_defunct_executed(self, mock_is_timed_out):
        self.task.state = types.TaskState.EXECUTED
        mock_is_timed_out.return_value = False
        self.assertFalse(self.task.is_defunct())
        mock_is_timed_out.assert_called_once_with(
            task_unresponsive_timeout_seconds
        )

    def test_defunct_executed(self, mock_is_timed_out):
        self.task.state = types.TaskState.EXECUTED
        mock_is_timed_out.return_value = True
        self.assertTrue(self.task.is_defunct())
        mock_is_timed_out.assert_called_once_with(
            task_unresponsive_timeout_seconds
        )

    def test_not_defunct_finished(self, mock_is_timed_out):
        self._assert_not_defunct(mock_is_timed_out, types.TaskState.FINISHED)


@mock.patch.object(tasks.Task, "_is_timed_out")
class TestAbandoned(TaskBaseTestCase):
    def _assert_not_abandoned(self, mock_is_timed_out, state):
        if self.task.state != state:
            self.task.state = state
        self.assertFalse(self.task.is_abandoned())
        mock_is_timed_out.assert_not_called()

    def test_not_abandoned_created(self, mock_is_timed_out):
        self._assert_not_abandoned(mock_is_timed_out, types.TaskState.CREATED)

    def test_not_abandoned_queued(self, mock_is_timed_out):
        self._assert_not_abandoned(mock_is_timed_out, types.TaskState.QUEUED)

    def test_not_abandoned_executed(self, mock_is_timed_out):
        self._assert_not_abandoned(mock_is_timed_out, types.TaskState.EXECUTED)

    def test_abandoned_finished(self, mock_is_timed_out):
        self.task.state = types.TaskState.FINISHED
        mock_is_timed_out.return_value = True
        self.assertTrue(self.task.is_abandoned())
        mock_is_timed_out.assert_called_once_with(
            task_abandoned_timeout_seconds
        )

    def test_not_abandoned_finished(self, mock_is_timed_out):
        self.task.state = types.TaskState.FINISHED
        mock_is_timed_out.return_value = False
        self.assertFalse(self.task.is_abandoned())
        mock_is_timed_out.assert_called_once_with(
            task_abandoned_timeout_seconds
        )



from .types import DeprecatedMessageCode as _D

# DEPRECATED REPORT CODES
# Move the unused report codes to the top of this file and remove
# the ReportItemMessage and CliReportMessageCustom definitions with associated
# report building tests

# Comment structure:
# Removed after <version>[, unused after <version>]

# Use known version number of last release. "Removed" means when was the report
# code deprecated. "Unused" means that the report is no longer produced by PCS
# but the report code and message are still defined.

# Optionally, if the report is being replaced by a different report, a comment
# with the new report code can be added


# Removed after 0.11.3
SBD_NOT_INSTALLED = _D("SBD_NOT_INSTALLED")
UNABLE_TO_DETERMINE_USER_UID = _D("UNABLE_TO_DETERMINE_USER_UID")
UNABLE_TO_DETERMINE_GROUP_GID = _D("UNABLE_TO_DETERMINE_GROUP_GID")

# Removed after 0.11.3, unused after 0.11.2
# Replaced by DEFAULTS_CAN_BE_OVERRIDDEN
# Fixed a typo in report code
DEFAULTS_CAN_BE_OVERRIDEN = _D("DEFAULTS_CAN_BE_OVERRIDEN")

# Removed after 0.11.3, unused after 0.10.10
# Replaced by ADD_REMOVE_* reports from new add/remove validator
CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_FOR_NEW_GROUP = _D(
    "CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_FOR_NEW_GROUP"
)
CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP = _D(
    "CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP"
)
CANNOT_GROUP_RESOURCE_ALREADY_IN_THE_GROUP = _D(
    "CANNOT_GROUP_RESOURCE_ALREADY_IN_THE_GROUP"
)
CANNOT_GROUP_RESOURCE_MORE_THAN_ONCE = _D(
    "CANNOT_GROUP_RESOURCE_MORE_THAN_ONCE"
)
CANNOT_GROUP_RESOURCE_NO_RESOURCES = _D("CANNOT_GROUP_RESOURCE_NO_RESOURCES")
CANNOT_GROUP_RESOURCE_NEXT_TO_ITSELF = _D(
    "CANNOT_GROUP_RESOURCE_NEXT_TO_ITSELF"
)

# Removed after 0.11.3, unused after 0.10.8
# Produced only with Pacemaker 1.x
MULTIPLE_SCORE_OPTIONS = _D("MULTIPLE_SCORE_OPTIONS")



from collections import (
    Counter,
    OrderedDict,
)
from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from lxml import etree
from lxml.etree import _Element

from pcs.common import reports
from pcs.common.reports import (
    ReportItem,
    ReportItemList,
)
from pcs.common.types import (
    StringCollection,
    StringIterable,
    StringSequence,
)
from pcs.lib.cib.resource.common import find_resources
from pcs.lib.cib.tools import (
    ElementSearcher,
    IdProvider,
    get_configuration_elements_by_id,
)
from pcs.lib.pacemaker.values import validate_id
from pcs.lib.xml_tools import (
    find_parent,
    move_elements,
    remove_one_element,
)

from .const import (
    TAG_OBJREF,
    TAG_TAG,
)


def _validate_tag_id(tag_id: str, id_provider: IdProvider) -> ReportItemList:
    """
    Validate that tag_id is a valid xml id and it is unique in the cib.

    tag_id -- identifier of new tag
    id_provider -- elements' ids generator
    """
    report_list: ReportItemList = []
    validate_id(tag_id, reporter=report_list)
    report_list.extend(id_provider.book_ids(tag_id))
    return report_list


def _validate_tag_id_not_in_idref_list(
    tag_id: str, idref_list: StringCollection
) -> ReportItemList:
    """
    Validate that idref_list does not contain tag_id.

    tag_id -- identifier of new tag
    idref_list -- reference ids which we want to tag
    """
    if tag_id in idref_list:
        return [ReportItem.error(reports.messages.TagCannotContainItself())]
    return []


def _validate_add_remove_duplicate_reference_ids(
    idref_list: StringSequence,
    add_or_not_remove: bool = True,
) -> ReportItemList:
    """
    Validate that idref_list does not contain duplicates.

    idref_list -- reference ids which we want to tag
    add_or_not_remove -- flag for add/remove action
    """
    duplicate_ids_list = [
        id for id, count in Counter(idref_list).items() if count > 1
    ]
    if duplicate_ids_list:
        return [
            ReportItem.error(
                reports.messages.TagAddRemoveIdsDuplication(
                    sorted(duplicate_ids_list),
                    add_or_not_remove,
                )
            )
        ]
    return []


def _validate_tag_create_idref_list_not_empty(
    idref_list: StringIterable,
) -> ReportItemList:
    """
    Validate that list of reference ids for tag create is not empty.

    idref_list -- reference ids to validate
    """
    # list for emptiness check (issue with some iterables like iterator)
    idref_list = list(idref_list)
    if not idref_list:
        return [
            ReportItem.error(
                reports.messages.TagCannotCreateEmptyTagNoIdsSpecified()
            )
        ]
    return []


def _validate_reference_ids_are_resources(
    resources_section: _Element,
    idref_list: StringSequence,
) -> ReportItemList:
    """
    Validate that ids are resources.

    resources_section -- element resources
    idref_list -- reference ids to validate
    """
    dummy_resources, report_list = find_resources(resources_section, idref_list)
    return report_list


def validate_create_tag(
    resources_section: _Element,
    tag_id: str,
    idref_list: StringSequence,
    id_provider: IdProvider,
) -> ReportItemList:
    """
    Validation function for tag creation.

    resources_section -- element resources
    tag_id -- identifier of new tag
    idref_list -- reference ids which we want to tag

    NOTE: Sequence vs. Collection issue:
            Value 'Collection' is unsubscriptable
            https://github.com/PyCQA/pylint/issues/2377
    """
    return (
        _validate_tag_id(tag_id, id_provider)
        + _validate_tag_create_idref_list_not_empty(idref_list)
        + _validate_tag_id_not_in_idref_list(tag_id, idref_list)
        + _validate_reference_ids_are_resources(resources_section, idref_list)
        + _validate_add_remove_duplicate_reference_ids(idref_list)
    )


def validate_remove_tag(
    constraint_section: _Element, to_remove_tag_list: StringIterable
) -> ReportItemList:
    """
    Validation function for tag removal. List of tag elements is not empty and
    no tag is referenced in any constraint.

    constraint_section -- element constraints
    to_remove_tag_list -- list of tag ids for removal
    """
    # list for emptiness check (issue with some iterables like iterator)
    to_remove_tag_list = list(to_remove_tag_list)
    if not to_remove_tag_list:
        return [
            ReportItem.error(
                reports.messages.TagCannotRemoveTagsNoTagsSpecified()
            )
        ]
    report_list = []
    for tag_id in to_remove_tag_list:
        constraint_list = find_constraints_referencing_tag(
            constraint_section,
            tag_id,
        )
        if constraint_list:
            report_list.append(
                ReportItem.error(
                    reports.messages.TagCannotRemoveTagReferencedInConstraints(
                        tag_id,
                        sorted(
                            [
                                str(elem.get("id", ""))
                                for elem in constraint_list
                            ]
                        ),
                    )
                )
            )
    return report_list


class ValidateTagUpdateByIds:
    # pylint:disable=too-many-instance-attributes
    """
    Validate update of a tag element by using ids.

    Class provides single validate method which returns a report item list,
    and methods for accessing required elements needed for update.

    Class is inspired by class ValidateMoveResourcesToGroupByIds from module
    'pcs/lib/cib/resource/hierarchy.py' which is used in command 'pcs resource
    group add'. It does validation based on given ids.  There is also class
    ValidateMoveResourcesToGroupByElements which implements validation based on
    given elements.
    Validation of tag update from existing elements is not implemented because
    it wasn't needed yet. If you need it, feel free to implement it. Take a look
    at the hierarchy module to see how it's done.
    """

    def __init__(
        self,
        tag_id: str,
        add_idref_list: StringSequence,
        remove_idref_list: StringSequence,
        adjacent_idref: Optional[str] = None,
    ) -> None:
        """
        tag_id -- id of an existing tag we want to update
        add_idref_list -- list of reference ids we want to add to the tag
        remove_idref_list -- list of reference ids we want to remove from the
            tag
        adjacent_idref -- id of an element next to which the added elements will
            be put
        """
        self._tag_id = tag_id
        self._add_idref_list = add_idref_list
        self._remove_idref_list = remove_idref_list
        self._adjacent_idref = adjacent_idref
        self._tag_element: Optional[_Element] = None
        self._add_obj_ref_element_list: List[_Element] = []
        self._adjacent_obj_ref_element: Optional[_Element] = None
        self._remove_obj_ref_element_list: List[_Element] = []

    def tag_element(self) -> Optional[_Element]:
        return self._tag_element

    def add_obj_ref_element_list(self) -> List[_Element]:
        return self._add_obj_ref_element_list

    def adjacent_obj_ref_element(self) -> Optional[_Element]:
        return self._adjacent_obj_ref_element

    def remove_obj_ref_element_list(self) -> List[_Element]:
        return self._remove_obj_ref_element_list

    def validate(
        self,
        resources_section: _Element,
        tags_section: _Element,
    ) -> ReportItemList:
        """
        Run the validation and return a report item list

        resources_section -- resources section of a cib
        tags_section -- tags section of a cib
        """
        return (
            self._validate_tag_exists(tags_section)
            + self._validate_ids_for_update_are_specified()
            + self._valdiate_no_common_add_remove_ids()
            + self._validate_ids_can_be_added_or_moved(resources_section)
            + self._validate_ids_can_be_removed()
            + self._validate_adjacent_id()
        )

    def _validate_tag_exists(self, tags_section: _Element) -> ReportItemList:
        """
        Validate that tag with given tag_id exists and save the found element.

        tags_section -- tags section of a cib
        """
        tag_list, report_list = find_tag_elements_by_ids(
            tags_section,
            [self._tag_id],
        )
        if tag_list:
            self._tag_element = tag_list[0]
        return report_list

    def _validate_ids_for_update_are_specified(self) -> ReportItemList:
        """
        Validate that either of add or remove ids were specified or that add ids
        were specified in case of specified adjacent id.
        """
        report_list: ReportItemList = []
        if not self._add_idref_list and not self._remove_idref_list:
            report_list.append(
                ReportItem.error(
                    reports.messages.TagCannotUpdateTagNoIdsSpecified()
                )
            )
        if not self._add_idref_list and self._adjacent_idref is not None:
            report_list.append(
                ReportItem.error(
                    reports.messages.TagCannotSpecifyAdjacentIdWithoutIdsToAdd(
                        self._adjacent_idref,
                    )
                )
            )
        return report_list

    def _valdiate_no_common_add_remove_ids(self) -> ReportItemList:
        """
        Validate that we do not remove ids currently being added.
        """
        common_ids = set(self._add_idref_list) & set(self._remove_idref_list)
        if common_ids:
            return [
                ReportItem.error(
                    reports.messages.TagCannotAddAndRemoveIdsAtTheSameTime(
                        sorted(common_ids),
                    )
                )
            ]
        return []

    def _validate_adjacent_id(self) -> ReportItemList:
        """
        Validate that element with adjacent id exists and it is not currently
        being added or removed.
        """
        report_list: ReportItemList = []
        if self._tag_element is None or self._adjacent_idref is None:
            return report_list
        self._adjacent_obj_ref_element = self._find_obj_ref_in_tag(
            self._adjacent_idref,
        )
        if self._adjacent_obj_ref_element is None:
            report_list.append(
                ReportItem.error(
                    reports.messages.TagAdjacentReferenceIdNotInTheTag(
                        self._adjacent_idref,
                        self._tag_id,
                    )
                )
            )
        if self._adjacent_obj_ref_element is not None:
            if self._adjacent_idref in self._add_idref_list:
                report_list.append(
                    ReportItem.error(
                        reports.messages.TagCannotPutIdNextToItself(
                            self._adjacent_idref,
                        )
                    )
                )
            if self._adjacent_idref in self._remove_idref_list:
                report_list.append(
                    ReportItem.error(
                        reports.messages.TagCannotRemoveAdjacentId(
                            self._adjacent_idref,
                        )
                    )
                )
        return report_list

    def _validate_ids_can_be_added_or_moved(
        self,
        resources_section: _Element,
    ) -> ReportItemList:
        """
        Validate that ids can be added or moved:
        - there are no duplicate ids specified
        - ids belong to elements allowed to be put in a tag in case of adding
        - ids do not exist in the tag if no adjacent id was specified
        Save a list of created (for ids newly added to the tag) and found (for
        ids already existing in the tag) obj_ref elements keeping the order of
        given ids to add or move them in case of success validation.
        """
        report_list: ReportItemList = []
        unique_add_ids = list(OrderedDict.fromkeys(self._add_idref_list))

        if self._add_idref_list:
            # report duplicate ids
            report_list.extend(
                _validate_add_remove_duplicate_reference_ids(
                    self._add_idref_list,
                )
            )
            # report if references not found or belongs to unexpected types
            report_list.extend(
                _validate_reference_ids_are_resources(
                    resources_section,
                    unique_add_ids,
                )
            )
        if self._tag_element is not None and self._add_idref_list:
            existing_element_id_list = []
            for id_ref in unique_add_ids:
                obj_ref = self._find_obj_ref_in_tag(id_ref)
                if obj_ref is None:
                    obj_ref = etree.Element(TAG_OBJREF, id=id_ref)
                else:
                    existing_element_id_list.append(id_ref)
                self._add_obj_ref_element_list.append(obj_ref)
            # report if a reference id exists in tag and no adjacent reference
            # id was specified
            if self._adjacent_idref is None and existing_element_id_list:
                report_list.append(
                    ReportItem.error(
                        # pylint: disable=line-too-long
                        reports.messages.TagCannotAddReferenceIdsAlreadyInTheTag(
                            self._tag_id,
                            sorted(existing_element_id_list),
                        )
                    )
                )
        return report_list

    def _validate_ids_can_be_removed(self) -> ReportItemList:
        """
        Validate that ids can be removed:
        - there are no duplicate ids specified
        - ids exist in the tag
        - tag would not be left empty after removal
        Saves found elements to remove in case of a successful validation.
        """
        report_list: ReportItemList = []
        if self._tag_element is None or not self._remove_idref_list:
            return report_list
        report_list.extend(
            _validate_add_remove_duplicate_reference_ids(
                self._remove_idref_list,
                add_or_not_remove=False,
            ),
        )
        missing_id_list = []
        for id_ref in set(self._remove_idref_list):
            obj_ref = self._find_obj_ref_in_tag(id_ref)
            if obj_ref is not None:
                self._remove_obj_ref_element_list.append(obj_ref)
            else:
                missing_id_list.append(id_ref)
        if missing_id_list:
            report_list.append(
                ReportItem.error(
                    reports.messages.TagIdsNotInTheTag(
                        self._tag_id,
                        sorted(missing_id_list),
                    )
                )
            )
        if not report_list and not self._add_idref_list:
            remove_difference = set(
                self._tag_element.findall(TAG_OBJREF),
            ).difference(set(self._remove_obj_ref_element_list))
            if not remove_difference:
                report_list.append(
                    ReportItem.error(
                        # pylint: disable=line-too-long
                        reports.messages.TagCannotRemoveReferencesWithoutRemovingTag(
                            self._tag_id,
                        )
                    )
                )
        return report_list

    def _find_obj_ref_in_tag(self, obj_ref_id: str) -> Optional[_Element]:
        """
        Find obj_ref element in the tag element being updated.

        obj_ref_id -- reference id
        """
        if self._tag_element is None:
            return None
        xpath_result = cast(
            List[_Element],
            self._tag_element.xpath(
                f"./{TAG_OBJREF}[@id=$obj_ref_id]", obj_ref_id=obj_ref_id
            ),
        )
        return xpath_result[0] if xpath_result else None


def find_constraints_referencing_tag(
    constraints_section: _Element,
    tag_id: str,
) -> List[_Element]:
    """
    Find constraint elements which are referencing specified tag.

    constraints_section -- element constraints
    tag_id -- tag id
    """
    constraint_list = constraints_section.xpath(
        """
        ./rsc_colocation[
            not (descendant::resource_set)
            and
            (@rsc=$tag_id or @with-rsc=$tag_id)
        ]
        |
        ./rsc_location[
            not (descendant::resource_set)
            and
            @rsc=$tag_id
        ]
        |
        ./rsc_order[
            not (descendant::resource_set)
            and
            (@first=$tag_id or @then=$tag_id)
        ]
        |
        ./rsc_ticket[
            not (descendant::resource_set)
            and
            @rsc=$tag_id
        ]
        |
        (./rsc_colocation|./rsc_location|./rsc_order|./rsc_ticket)[
            ./resource_set/resource_ref[@id=$tag_id]
        ]
        """,
        tag_id=tag_id,
    )
    return cast(List[_Element], constraint_list)


def find_tag_elements_by_ids(
    tags_section: _Element,
    tag_id_list: StringIterable,
) -> Tuple[List[_Element], ReportItemList]:
    """
    Try to find tag elements by ids and return them with non-empty report
    list in case of errors.

    tags_section -- element tags
    tag_id_list -- list of tag indentifiers
    """
    element_list = []
    report_list: ReportItemList = []
    for tag_id in tag_id_list:
        searcher = ElementSearcher(TAG_TAG, tag_id, tags_section)
        if searcher.element_found():
            element_list.append(searcher.get_element())
        else:
            report_list.extend(searcher.get_errors())

    return element_list, report_list


def create_tag(
    tags_section: _Element, tag_id: str, idref_list: StringIterable
) -> _Element:
    """
    Create new tag element and add it to cib.
    Returns newly created tag element.

    tags_section -- element tags
    tag_id -- identifier of new tag
    idref_list -- reference ids which we want to tag
    """
    tag_el = etree.SubElement(tags_section, TAG_TAG, id=tag_id)
    for ref_id in idref_list:
        etree.SubElement(tag_el, TAG_OBJREF, id=ref_id)
    return tag_el


def remove_tag(
    tag_elements: Iterable[_Element],
) -> None:
    """
    Remove given tag elements from a cib.

    tag_elements -- tag elements to be removed
    """
    for tag in tag_elements:
        remove_one_element(tag)


def remove_obj_ref(obj_ref_list: Iterable[_Element]) -> None:
    """
    Remove specified obj_ref elements and also their parents if they remain
    empty after obj_ref removal.

    obj_ref_list -- list of obj_ref elements
    """
    tag_elements = {
        element
        for element in {
            find_parent(obj_ref, [TAG_TAG]) for obj_ref in obj_ref_list
        }
        if element is not None
    }

    for obj_ref in obj_ref_list:
        remove_one_element(obj_ref)
    for tag in tag_elements:
        if len(tag.findall(TAG_OBJREF)) == 0:
            remove_one_element(tag)


def add_obj_ref(
    tag_element: _Element,
    obj_ref_el_list: Iterable[_Element],
    adjacent_element: Optional[_Element],
    put_after_adjacent: bool = False,
) -> None:
    """
    Add or move obj_ref elements in given tag element.

    tag_element -- the tag element to be updated with obj_ref elements
    obj_ref_el_list -- elements to be added or moved in the updated tag
    adjacent_element -- the element next to which the obj_ref elements will be
        put
    put_after_adjacent -- put elements after (True) or before (False) the
        adjacent element
    """
    if adjacent_element is not None:
        move_elements(
            obj_ref_el_list,
            adjacent_element,
            put_after_adjacent=put_after_adjacent,
        )
    else:
        for obj_ref in obj_ref_el_list:
            tag_element.append(obj_ref)


def get_list_of_tag_elements(tags_section: _Element) -> List[_Element]:
    """
    Get list of tag elements from cib.

    tags_section -- element tags
    """
    return tags_section.findall(TAG_TAG)


def tag_element_to_dict(
    tag_element: _Element,
) -> dict[str, Union[str, list[str]]]:
    """
    Convert tag element to the dict structure

    tag_element -- single tag element that contains obj_ref elements

    {
        "tag_id": "tag1",
        "idref_list": ["i1", "i2"],
    },
    """
    return {
        # NOTE: .get("id", default="") for typing  there always be an id
        "tag_id": str(tag_element.get("id", default="")),
        "idref_list": [
            str(obj_ref.get("id", default=""))
            for obj_ref in tag_element.findall(TAG_OBJREF)
        ],
    }


def expand_tag(
    some_or_tag_el: _Element,
    only_expand_types: Optional[StringCollection] = None,
) -> List[_Element]:
    """
    Substitute a tag element with elements which the tag refers to.

    some_or_tag_el -- an already expanded element or a tag element to expand
    only_expand_types -- if specified, return only elements of these types
    """
    if some_or_tag_el.tag != TAG_TAG:
        return [some_or_tag_el]

    conf_section = find_parent(some_or_tag_el, {"configuration"})
    if conf_section is None:
        return []

    expanded_elements = []
    for element_id in [
        str(obj_ref.get("id", ""))
        for obj_ref in some_or_tag_el.iterfind(TAG_OBJREF)
    ]:
        if only_expand_types:
            searcher = ElementSearcher(
                only_expand_types, element_id, conf_section
            )
            if searcher.element_found():
                expanded_elements.append(searcher.get_element())
        else:
            expanded_elements.extend(
                get_configuration_elements_by_id(conf_section, element_id)
            )
    return expanded_elements



import re
from collections import namedtuple

from pcs.common import reports
from pcs.common.reports import (
    ReportItem,
    get_severity,
)
from pcs.lib.booth import constants
from pcs.lib.interface.config import (
    ExporterInterface,
    ParserErrorException,
    ParserInterface,
)


class ConfigItem(namedtuple("ConfigItem", "key value details")):
    def __new__(cls, key, value, details=None):
        return super().__new__(cls, key, value, details or [])


class InvalidLines(ParserErrorException):
    pass


class Parser(ParserInterface):
    @staticmethod
    def parse(raw_file_data):
        return _organize_lines(
            _parse_to_raw_lines(raw_file_data.decode("utf-8"))
        )

    @staticmethod
    def exception_to_report_list(
        exception, file_type_code, file_path, force_code, is_forced_or_warning
    ):
        del file_type_code  # this is defined by the report code
        if isinstance(exception, InvalidLines):
            return [
                ReportItem(
                    severity=get_severity(force_code, is_forced_or_warning),
                    message=reports.messages.BoothConfigUnexpectedLines(
                        exception.args[0],
                        file_path,
                    ),
                )
            ]
        raise exception


class Exporter(ExporterInterface):
    @staticmethod
    def export(config_structure):
        return "\n".join(_build_to_lines(config_structure) + [""]).encode(
            "utf-8"
        )


def _build_to_lines(config_line_list, deep=0):
    line_list = []
    for key, value, details in config_line_list:
        line_value = value if key != "ticket" else '"{0}"'.format(value)
        line_list.append("{0}{1} = {2}".format("  " * deep, key, line_value))
        if details:
            line_list.extend(_build_to_lines(details, deep + 1))
    return line_list


def _organize_lines(raw_line_list):
    # Decision: Global key is moved up when is below ticket. Alternative is move
    # it below all ticket details. But it is confusing.
    global_section = []
    ticket_section = []
    current_ticket = None
    for key, value in raw_line_list:
        if key == "ticket":
            current_ticket = ConfigItem(key, value)
            ticket_section.append(current_ticket)
        elif key in constants.GLOBAL_KEYS or not current_ticket:
            global_section.append(ConfigItem(key, value))
        else:
            current_ticket.details.append(ConfigItem(key, value))

    return global_section + ticket_section


def _search_with_multiple_re(re_object_list, string):
    """
    return MatchObject of first matching regular expression object or None
    list re_object_list contains regular expression objects (products of
        re.compile)
    """
    for expression in re_object_list:
        match = expression.search(string)
        if match:
            return match
    return None


def _parse_to_raw_lines(config_content):
    keyword_part = r"^(?P<key>[a-zA-Z0-9_-]+)\s*=\s*"
    expression_list = [
        re.compile(pattern.format(keyword_part))
        for pattern in [
            r"""{0}(?P<value>[^'"]+)$""",
            r"""{0}'(?P<value>[^']*)'\s*(#.*)?$""",
            r"""{0}"(?P<value>[^"]*)"\s*(#.*)?$""",
        ]
    ]

    line_list = []
    invalid_line_list = []
    for line in config_content.splitlines():
        line = line.strip()
        match = _search_with_multiple_re(expression_list, line)
        if match:
            line_list.append((match.group("key"), match.group("value")))
        elif line and not line.startswith("#"):
            invalid_line_list.append(line)

    if invalid_line_list:
        raise InvalidLines(invalid_line_list)

    return line_list



from logging import Logger
from unittest import (
    TestCase,
    mock,
)

from pcs.common.file import (
    FileMetadata,
    RawFileError,
)
from pcs.common.file_type_codes import PCS_SETTINGS_CONF
from pcs.lib.auth.const import (
    ADMIN_GROUP,
    SUPERUSER,
)
from pcs.lib.auth.types import AuthUser
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.json import JsonParserException
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.permissions.checker import PermissionsChecker
from pcs.lib.permissions.config.facade import FacadeV2
from pcs.lib.permissions.config.parser import ParserError
from pcs.lib.permissions.config.types import (
    ClusterPermissions,
    ConfigV2,
    PermissionAccessType,
    PermissionEntry,
    PermissionTargetType,
)

_FILE_PATH = "file path"
_FILE_METADATA = FileMetadata(
    file_type_code=PCS_SETTINGS_CONF,
    path=_FILE_PATH,
    owner_user_name="root",
    owner_group_name="root",
    permissions=None,
    is_binary=False,
)


def _config_fixture(permissions=tuple()):
    return ConfigV2(
        data_version=1,
        clusters=[],
        permissions=ClusterPermissions(local_cluster=permissions),
    )


_FACADE_FIXTURE = FacadeV2(
    _config_fixture(
        (
            PermissionEntry(
                name="user-read",
                type=PermissionTargetType.USER,
                allow=(PermissionAccessType.READ,),
            ),
            PermissionEntry(
                name="user-write",
                type=PermissionTargetType.USER,
                allow=(PermissionAccessType.WRITE,),
            ),
            PermissionEntry(
                name="user-full",
                type=PermissionTargetType.USER,
                allow=(PermissionAccessType.FULL,),
            ),
            PermissionEntry(
                name="group-grant",
                type=PermissionTargetType.GROUP,
                allow=(PermissionAccessType.GRANT,),
            ),
        )
    )
)


_EMPTY_CONFIG = _config_fixture()


class PermissionCheckerGetFacadeTest(TestCase):
    # pylint: disable=protected-access
    def setUp(self):
        self.file_instance_mock = mock.Mock(spec_set=FileInstance)
        self.file_instance_mock.raw_file.metadata = _FILE_METADATA
        self.file_instance_mock.raw_file.exists.return_value = True
        self.logger = mock.Mock(spec_set=Logger)
        with mock.patch.object(
            FileInstance,
            "for_pcs_settings_config",
            lambda *_args, **_kwargs: self.file_instance_mock,
        ):
            self.provider = PermissionsChecker(self.logger)

    def test_io_error(self):
        reason = "reason"
        self.file_instance_mock.read_to_facade.side_effect = RawFileError(
            _FILE_METADATA, RawFileError.ACTION_READ, reason
        )
        self.assertEqual(_EMPTY_CONFIG, self.provider._get_facade().config)
        self.logger.error.assert_called_once_with(
            "Unable to read file '%s': %s", _FILE_PATH, reason
        )

    def test_file_doesnt_exist(self):
        reason = "reason"
        self.file_instance_mock.raw_file.exists.return_value = False
        self.file_instance_mock.read_to_facade.side_effect = RawFileError(
            _FILE_METADATA, RawFileError.ACTION_READ, reason
        )
        self.assertEqual(
            ConfigV2(
                data_version=1,
                clusters=[],
                permissions=ClusterPermissions(
                    local_cluster=(
                        PermissionEntry(
                            type=PermissionTargetType.GROUP,
                            name=ADMIN_GROUP,
                            allow=(
                                PermissionAccessType.READ,
                                PermissionAccessType.WRITE,
                                PermissionAccessType.GRANT,
                            ),
                        ),
                    ),
                ),
            ),
            self.provider._get_facade().config,
        )
        self.logger.debug.assert_called_once_with(
            "File '%s' doesn't exist, using default configuration", _FILE_PATH
        )

    def test_json_parser_error(self):
        self.file_instance_mock.read_to_facade.side_effect = (
            JsonParserException(None)
        )
        self.assertEqual(_EMPTY_CONFIG, self.provider._get_facade().config)
        self.logger.error.assert_called_once_with(
            "Unable to parse file '%s': not valid json", _FILE_PATH
        )

    def test_invalid_format(self):
        reason = "reason"
        self.file_instance_mock.read_to_facade.side_effect = ParserError(reason)
        self.assertEqual(_EMPTY_CONFIG, self.provider._get_facade().config)
        self.logger.error.assert_called_once_with(
            "Unable to parse file '%s': %s", _FILE_PATH, reason
        )

    def test_other_parsing_error(self):
        self.file_instance_mock.read_to_facade.side_effect = (
            ParserErrorException()
        )
        self.assertEqual(_EMPTY_CONFIG, self.provider._get_facade().config)
        self.logger.error.assert_called_once_with(
            "Unable to parse file '%s'", _FILE_PATH
        )

    def test_success(self):
        facade = FacadeV2(
            ConfigV2(
                data_version=1,
                clusters=[],
                permissions=ClusterPermissions(
                    local_cluster=(
                        PermissionEntry(
                            name="user1",
                            type=PermissionTargetType.USER,
                            allow=(PermissionAccessType.WRITE),
                        ),
                    )
                ),
            )
        )
        self.file_instance_mock.read_to_facade.return_value = facade
        self.assertEqual(facade, self.provider._get_facade())
        self.logger.error.assert_not_called()


@mock.patch.object(
    PermissionsChecker, "_get_facade", lambda _self: _FACADE_FIXTURE
)
class PermissionCheckerGetPermissions(TestCase):
    def setUp(self):
        self.logger = mock.Mock(spec_set=Logger)
        self.checker = PermissionsChecker(self.logger)

    def test_superuser(self):
        self.assertEqual(
            set(PermissionAccessType),
            self.checker.get_permissions(
                AuthUser(username=SUPERUSER, groups=tuple())
            ),
        )

    def test_completion_full(self):
        self.assertEqual(
            {
                PermissionAccessType.READ,
                PermissionAccessType.WRITE,
                PermissionAccessType.GRANT,
                PermissionAccessType.FULL,
            },
            self.checker.get_permissions(
                AuthUser(username="user-full", groups=tuple())
            ),
        )

    def test_completion_write(self):
        self.assertEqual(
            {
                PermissionAccessType.READ,
                PermissionAccessType.WRITE,
            },
            self.checker.get_permissions(
                AuthUser(username="user-write", groups=tuple())
            ),
        )

    def test_groups_permissions(self):
        self.assertEqual(
            {PermissionAccessType.GRANT},
            self.checker.get_permissions(
                AuthUser(username="user", groups=("group-grant",))
            ),
        )

    def test_user_and_groups_permissions(self):
        self.assertEqual(
            {
                PermissionAccessType.READ,
                PermissionAccessType.WRITE,
                PermissionAccessType.GRANT,
            },
            self.checker.get_permissions(
                AuthUser(username="user-write", groups=("group-grant",))
            ),
        )


@mock.patch.object(
    PermissionsChecker, "_get_facade", lambda _self: _FACADE_FIXTURE
)
class PermissionsCheckerIsAuthorizedTest(TestCase):
    def setUp(self):
        self.logger = mock.Mock(spec_set=Logger)
        self.checker = PermissionsChecker(self.logger)

    def test_allowed(self):
        user = AuthUser("user-full", ("group1", "group2"))
        access = PermissionAccessType.READ
        self.assertTrue(self.checker.is_authorized(user, access))
        self.assertEqual(
            [
                mock.call(
                    "Permission check: username=%s groups=%s access=%s",
                    user.username,
                    ",".join(user.groups),
                    str(access.value),
                ),
                mock.call(
                    "Current user permissions: %s",
                    ",".join(
                        sorted(
                            str(permission.value)
                            for permission in (
                                PermissionAccessType.READ,
                                PermissionAccessType.WRITE,
                                PermissionAccessType.GRANT,
                                PermissionAccessType.FULL,
                            )
                        )
                    ),
                ),
                mock.call("%s access granted", str(access.value).capitalize()),
            ],
            self.logger.debug.mock_calls,
        )

    def test_not_allowed(self):
        user = AuthUser("user-full", ("group1", "group2"))
        access = PermissionAccessType.SUPERUSER
        self.assertFalse(self.checker.is_authorized(user, access))
        self.assertEqual(
            [
                mock.call(
                    "Permission check: username=%s groups=%s access=%s",
                    user.username,
                    ",".join(user.groups),
                    str(access.value),
                ),
                mock.call(
                    "Current user permissions: %s",
                    ",".join(
                        sorted(
                            str(permission.value)
                            for permission in (
                                PermissionAccessType.READ,
                                PermissionAccessType.WRITE,
                                PermissionAccessType.GRANT,
                                PermissionAccessType.FULL,
                            )
                        )
                    ),
                ),
                mock.call("%s access denied", str(access.value).capitalize()),
            ],
            self.logger.debug.mock_calls,
        )



import errno
import os
from contextlib import contextmanager
from io import BytesIO
from typing import (
    Dict,
    Iterator,
    Optional,
)

# the import makes it look like RealFile is implemented here so we don't
# have to import RawFile from common and Ghost file from here in other
# places
# pylint: disable=unused-import
from pcs.common import reports
from pcs.common.file import FileMetadata
from pcs.common.file import RawFile as RealFile
from pcs.common.file import (
    RawFileError,
    RawFileInterface,
)

# TODO add logging (logger / debug reports ?)


def raw_file_error_report(
    error: RawFileError,
    force_code: Optional[reports.types.ForceCode] = None,
    is_forced_or_warning: bool = False,
) -> reports.ReportItem:
    """
    Translate a RawFileError instance to a report

    error -- an exception to be translated
    force_code -- is it a forcible error? by which code?
    is_forced_or_warning -- translate to a warning if True, error otherwise
    """
    return reports.ReportItem(
        severity=reports.get_severity(force_code, is_forced_or_warning),
        message=reports.messages.FileIoError(
            error.metadata.file_type_code,
            error.action,
            error.reason,
            # do not report real file path if we were working with a ghost file
            file_path=(
                "" if isinstance(error, GhostFileError) else error.metadata.path
            ),
        ),
    )


class GhostFileError(RawFileError):
    pass


class GhostFile(RawFileInterface):
    def __init__(
        self, metadata: FileMetadata, file_data: Optional[bytes] = None
    ):
        """
        metadata -- describes the file and provides its metadata
        file_data -- data of the ghost file
        """
        super().__init__(metadata)
        self.__file_data = file_data

    @property
    def content(self) -> Optional[bytes]:
        """
        Export the file content
        """
        return self.__file_data

    def exists(self) -> bool:
        return self.__file_data is not None

    def read(self) -> bytes:
        if self.__file_data is None:
            raise GhostFileError(
                self.metadata,
                RawFileError.ACTION_READ,
                # get "no such file" message as defined and worded in the system
                os.strerror(errno.ENOENT),
            )
        return self.__file_data

    def write(self, file_data: bytes, can_overwrite: bool = False) -> None:
        self.__file_data = file_data

    @contextmanager
    def update(self) -> Iterator[BytesIO]:
        io_buffer = BytesIO(self.read())
        io_buffer.seek(0)
        yield io_buffer
        self.write(io_buffer.getvalue(), can_overwrite=True)


def export_ghost_file(ghost_file: GhostFile) -> Dict[str, Optional[bytes]]:
    """
    Export GhostFile so it can be transferred to a client

    ghost_file -- a ghost file instance
    """
    return {
        "content": ghost_file.content,
    }
